from .variable import Variable


def replace_variables(d, data: dict):
    """
    Replace variables in `d` (usually BaseQuery.query)
    with data from the `data` dict.
    """
    if type(d) is dict:
        d = d.copy()
        for k, v in d.items():
            if isinstance(k, Variable):
                d[k.value_from_dict(data)] = d[k]
                del d[k]
                k = k.value_from_dict(data)

            if isinstance(v, Variable):
                d[k] = v.value_from_dict(data)
            else:
                d[k] = replace_variables(v, data)
    elif type(d) is list:
        d = d[:]
        for i, v in enumerate(d):
            if isinstance(v, Variable):
                d[i] = v.value_from_dict(data)
            else:
                d[i] = replace_variables(v, data)
    elif isinstance(d, Variable):
        d = d.value_from_dict(data)
    return d



from .query import BaseQuery


class QueryFilter(object):
    """
    A QueryFilter is an object that can be used in a query to restrict
    appearances of certain elements.

    Usually, a QueryFilter will override the `apply()` method and will
    change the received query into what's needed to filter further. As
    such, a filter exports a variables dictionary, with additional
    data which may (or must) be received from the API call. A filter
    may be required or optional (optional filters makes all its
    variables optional).

    The variable query_field is used to define which part of the query
    will be filtered (query, aggs, size, sort)
    """
    query_field = 'query'
    variables = {}
    required = False

    def __call__(self, query: BaseQuery, data={}):
        """
        A filter is callable this way. Receives the query (which can be
        mutated) and the data from forms or otherwise.
        """
        if self.can_apply(data):
            return self.apply(query, data)
        else:
            if self.required:
                raise Exception(
                    'Filter {} is required but variables were missing.'.format(
                        self.__name__))
            return query

    def get_variables(self):
        if self.required:
            return {
                k: variable.copy()
                for k, variable in self.variables.items()
            }
        else:
            return {
                k: variable.copy(required=False)
                for k, variable in self.variables.items()
            }

    def can_apply(self, data):
        """
        Whether we can apply the filter, notwithstanding its required property.

        A filter can be applied if all its required variables are found in
        `data`.

        """

        for k, v in self.variables.items():
            if v.name not in data:
                return False
        return True

    def apply(self, query, data):
        """
        Override this function in your filter so that it does something.
        """
        return query



# -*- coding: utf-8 -*-
from pbr.version import VersionInfo
_v = VersionInfo(__name__).semantic_version()

__author__ = """Santiago Saavedra"""
__email__ = 'ssaavedra@openshine.com'

__version__ = _v.release_string()
version_info = _v.version_tuple()



import json

from .variable import Variable

__all__ = ['generate_query_docs']

_docs_template = """
* **ID**: ``{id}``
* **name**: ``{name}``

{long_description}

.. code-block:: json

  {query}
"""

_view_template = """

* **Query**:

.. code-block:: json

    {query}

    """


def generate_query_docs(q, variables=None):
    """
    Return a full description of a the query to create the documentation.
    """
    return _docs_template.format(
        id=q.get_id(),
        name=q.name,
        long_description=q.__doc__ or '',
        query='\n  '.join(
            json.dumps(
                q.get_es_query(variables or {}),
                indent=2,
                default=_explaining_json_encoder,
                sort_keys=True).split('\n')))


def generate_view_docs(q, variables=None):
    """
    Return a simple description of a the query to create the view documentation.
    """
    return _view_template.format(
            query='\n  '.join(
                json.dumps(
                    q.get_es_query(variables or {}),
                    indent=2,
                    default=_explaining_json_encoder,
                    sort_keys=True).split('\n')))


def _explaining_json_encoder(obj):
    """
    A JSON encoder which understands Variables.
    """
    if isinstance(obj, Variable):
        return str(obj)
    else:
        return obj



__all__ = ['Variable']


class Variable(object):
    """A Variable is an element to be embedded in a query.  The variable
    can have a default value (if it is not required), and it has a
    help text, as well as a type. From this variable, a DRF serializer
    can be constructed. The serializer class can be automatically
    selected, but may be overridden in the cosntructor, as well as
    further parameters.
    You can define a method (builder) that builds a part of the query
    if you need more than a simple replacement.
    For example, build a date_histogram or a histogram depending on the value.
    """

    def __init__(
            self,
            name: str,
            default=None,
            type: type=str,
            required: bool=False,
            help_text: str='Unknown variable',
            serializer_class=None,
            serializer_options=None,
            builder=None):
        self.name = name
        self.default = default
        self.type = type
        self.required = required
        self.help_text = help_text
        self.serializer_options = serializer_options or {}
        self.serializer_class = serializer_class
        self.builder = builder

    def copy(self, **kwargs):
        """
        Copy the variable, optionally changing some of the values.

        Implementation nodte: currently it uses the __dict__ of the
        object for the copy.
        """
        d = self.__dict__.copy()
        d.update(kwargs)
        return Variable(**d)

    def value_from_dict(self, d: dict=None):
        """
        Returns the value of the variable given the received data.

        If no data matches this variable name, its default value is
        returned.

        """
        if d is None:
            d = {}
        if self.required and self.name not in d:
            raise Exception(
                "Required variable {} does not have a value".format(self.name))
        else:
            if self.builder is None:
                return d.get(self.name, self.default)
            return self.builder(d.get(self.name, self.default))

    def __str__(self):
        """
        Returns a formatted text of this variable as an explanation.
        """
        if self.required:
            return '${{{name}:?Required. {doc}}}'.format(
                name=self.name,
                doc=self.help_text, )
        else:
            return '${{{name}:-{default}:?{doc}}}'.format(
                name=self.name,
                default=self.default,
                doc=self.help_text, )

    def __repr__(self):
        """Returns a representation of the object for schematic purposes.
        (R) means required, (O) means optional.
        """
        return "<Var [{name} :: {type} = {default} {req}]>".format(
            name=self.name,
            type=self.type.__name__,
            default=repr(self.default),
            req="(R)" if self.required else "(O)")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_esqb
----------------------------------

Tests for `esqb` module.
"""

from esqb import query, variable

testvar = variable.Variable('t')


class Query(query.BaseQuery):
    aggs = {
        'x': testvar
    }


def test_contains_var():
    assert testvar in Query().find_all_variables()


def test_can_override_var():
    q = Query().get_es_query({'t': 'v'})
    assert q['aggs'] == {'x': 'v'}
    assert q['query'] == {}
    assert q['size'] == 0


def test_uses_default_value():
    class RQuery(query.BaseQuery):
        aggs = {
            'x': variable.Variable(
                't', type=bool,
                required=False,
                default=True)
        }

    q = RQuery()
    assert q.get_es_query({})['aggs'] == {'x': True}


def test_throws_on_missing_required():
    class RQuery(query.BaseQuery):
        aggs = {
            'x': variable.Variable(
                't', type=bool,
                required=True
            )
        }

    q = RQuery()
    try:
        assert q.get_es_query({})
    except Exception:
        pass


def test_complex_query_contains_multiple_vars():
    class RQuery(query.BaseQuery):
        aggs = {
            'must': {
                'bool': [
                    {
                        'match': {
                            variable.Variable(
                                'term', default=0
                            ): variable.Variable('value', default=0)
                        }
                    },
                ]
            }
        }

    q = RQuery()

    assert q.get_es_query({'term': 'x', 'value': 'y'})['aggs'] == {
        'must': {
            'bool': [
                {
                    'match': {
                        'x': 'y'
                    }
                }
            ]
        }
    }



import os
import sys
from copy import deepcopy

from .variable import Variable
from .utils import replace_variables as _replace_variables


class BaseQuery(object):
    """
    This object contains everything necessary to create serializers
    from queries. This object resolves variables in queries as well as
    possible filters (optional or required) which may change the final
    query to be given to ES.

    """

    def __init__(self):
        """
        We make this arcane magic here so that we can have several
        instances of the same query and their base (class-based)
        attributes will not change if modified (because there is a
        copy in the instance).

        The serializer field exists to cache a serializer once it has
        been generated.

        """
        self._size = self.__class__.__dict__.get('size', 0)
        self._aggs = deepcopy(self.__class__.__dict__.get('aggs', {}))
        self._query = deepcopy(self.__class__.__dict__.get('query', {}))
        self.filters = deepcopy(self.__class__.__dict__.get('filters', []))
        self._sort = deepcopy(self.__class__.__dict__.get('sort', []))
        self._serializer = None

    def get_id(self):
        """
        TODO: Probably just return self.__class__.__module__
        """
        fullpath = sys.modules[self.__class__.__module__].__file__
        name = os.path.basename(fullpath)
        dot = name.rfind('.')
        return (name[:dot], fullpath)

    @property
    def aggs(self):
        """
        Returns the ``aggs`` property of the query.
        """
        return self._aggs

    @aggs.setter
    def aggs(self, val: dict):
        self._aggs = val

    @property
    def query(self):
        """
        Returns the ``query`` property of the query.
        """
        return self._query

    @query.setter
    def query(self, val: dict):
        self._query = val

    @property
    def size(self):
        """
        Returns the ``size`` property of the query.
        """
        return self._size

    @size.setter
    def size(self, val):
        self._size = val

    @property
    def sort(self):
        """
        Returns the ``sort`` property of the query.
        """
        return self._sort

    @sort.setter
    def sort(self, val):
        self._sort = val

    def get_es_query(self, data: dict) -> dict:
        """
        Return the final ES query that should be presented to ESService.

        This function scans variables and fills their values with the
        provided arg `data`.
        """
        _vars = {}
        for var in self.find_all_variables():
            _vars[var.name] = var

        return {_q: _replace_variables(self._filtered(_q, data), data)
                for _q in ('query', 'size', 'aggs', 'sort')}

    def _filtered(self, query_field, data: dict) -> dict:
        """Filters the query parts with the defined query filters"""
        d = deepcopy(getattr(self, query_field))
        for _filter in self.filters:
            if _filter.query_field == query_field:
                d = _filter(d, data)
        return d

    @property
    def serializer(self, **kwargs):
        """
        Creates a serializer instance and returns such an instance on
        following calls.

        """
        if not self._serializer:
            self._serializer = self.get_serializer()(**kwargs)
        return self._serializer

    @serializer.setter
    def serializer(self, value):
        self._serializer = value

    def get_serializer(self):
        """
        You should use the utility function in drf_support so that
        this module becomes drf_free.
        """
        import warnings
        warnings.warn("Query.get_serialiser is deprecated. " +
                      "Use drf_support.get_query_serializer instead(query)",
                      DeprecationWarning)
        from . import drf_support
        return drf_support.get_query_serializer(self)

    def find_all_variables(self, docs=None, prev=None, include_filters=True):
        """
        Returns all variables found in the query and its filters
        """
        if docs is None:
            prev = [] if not include_filters else [
                var
                for _filter in self.filters
                for var in _filter.get_variables().values() if include_filters
            ]
            return self.find_all_variables(
                [self.query, self.aggs, self.size, self.sort], prev
            )

        if type(docs) is list:
            for elt in docs:
                if type(elt) is Variable:
                    prev.append(elt)
                else:
                    self.find_all_variables(elt, prev)
        elif type(docs) is dict:
            for k, v in docs.items():
                if type(k) is Variable:
                    prev.append(k)
                if type(v) is Variable:
                    prev.append(v)
                else:
                    self.find_all_variables(v, prev)
        return prev

    def dotget(self, doc: dict, path: str):
        """
        Utility function to navigate paths using dot-notation.
        """
        for key in path.split('.'):
            doc = doc[key]
        return doc

    @property
    def name(self):
        """
        Unless overwritten, the name is the class name
        """
        return self.__class__.__name__

    @property
    def full_query(self):
        """
        Returns the query as it would be compiled to be sent to
        ElasticSearch.

        Writing to this property is not supported at the
        moment. Please use ``aggs`` and ``query``.

        """
        return {
            'size': self.size,
            'query': self.query,
            'aggs': self.aggs,
            'sort': self.sort
        }

    def docs(self, variables):
        """
        Returns the documentation of this query.

        This method is deprecated, please use
        :func:`esqb.docs_builder.generate_query_docs`

        """
        import warnings
        warnings.warn('query.docs is deprecated. ' +
                      'Use docs_builder.generate_query_docs(query)',
                      DeprecationWarning)
        from .docs_builder import generate_query_docs
        return generate_query_docs(self, variables)

    def view_docs(self, variables):
        """
        Returns the view documentation of this query.

        This method is deprecated, please use
        :func:`esqb.docs_builder.generate_view_docs`

        """
        import warnings
        warnings.warn('query.docs is deprecated. ' +
                      'Use docs_builder.generate_view_docs(query)',
                      DeprecationWarning)
        from .docs_builder import generate_view_docs
        return generate_view_docs(self, variables)



import collections

from django.core.exceptions import ImproperlyConfigured
try:
    from rest_framework import serializers
except ImproperlyConfigured:
    import os
    import warnings
    warnings.warn("""Django settings not detected.
You should use drf_support inside a Django application.
Continuing with an empty DJANGO_SETTINGS_MODULE
""")
    os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'x')
    from rest_framework import serializers
from rest_framework.utils import html


def get_variable_serializer_field(variable):
    """"
    Returns a serializer field for a variable.

    The variable MAY add parameters to the serializer via a
    serializer_field_options property.

    """
    if hasattr(variable, 'serializer_field_options'):
        sfo = variable.serializer_field_options or {}
    else:
        sfo = {}

    sfo.setdefault('required', variable.required)
    if not variable.required and variable.default:
        sfo['default'] = variable.default
    elif 'default' in sfo:
        del sfo['default']

    return get_variable_serializer_field_class(variable)(
        help_text=variable.help_text, **sfo)


def get_variable_serializer_field_class(variable):
    """
    Usually, you should not worry about this and use
    get_query_serializer directly.

    But you may add a variable.serializer_class parameter if you wish
    to customize the serializer field that your variable outputs.

    """
    if hasattr(variable, 'serializer_field_class') and \
       variable.serializer_field_class:
        return variable.serializer_field_class

    return _default_variable_serializers[variable.type]


def get_query_serializer(query, hide=()):
    """Creates a serializer synthetically with parameters defined by all
    the variables defined in the query and its available filters.

    You can avoid pushing variables to the serializer if you
    include them (either strings containing their names or the
    objects themselves) in the optional hide parameter.

    """
    name = "Synthetic" + query.get_id()[0] + "Serializer"
    params = {
        var.name: get_variable_serializer_field(var)
        for var in query.find_all_variables()
        if var not in hide and var.name not in hide
    }
    return type(name, (serializers.Serializer, ), params)


class ListField(serializers.ListField):
    def to_internal_value(self, data):
        """
        List of dicts of native values <- List of dicts of primitive datatypes.
        """
        _data = data[0].split(',') if len(data) > 0 else data
        if html.is_html_input(_data):
            data = html.parse_html_list(_data)
        if isinstance(_data, type('')) or \
           isinstance(_data, collections.Mapping) or \
           not hasattr(data, '__iter__'):
            self.fail('not_a_list', input_type=type(_data).__name__)
        if not self.allow_empty and len(_data) == 0:
            self.fail('empty')
        return [self.child.run_validation(item) for item in _data]


_default_variable_serializers = {
    int: serializers.IntegerField,
    float: serializers.FloatField,
    str: serializers.CharField,
    bool: serializers.BooleanField,
    list: ListField
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_esqb
----------------------------------

Tests for `esqb` module.
"""

from esqb import queryfilter, query, variable

testvar = variable.Variable('t')


def test_queryfilter_adds_optional_vars_if_filter_optional_and_vars_required():
    class QF(queryfilter.QueryFilter):
        required = False
        variables = {
            't': variable.Variable('t', required=True)
        }

        def apply(self, query):
            return query

    class Q(query.BaseQuery):
        filters = [QF()]

    assert len(Q().find_all_variables()) == 1
    assert Q().find_all_variables()[0].required is False


def test_queryfilter_adds_required_vars_if_filter_required_and_vars_required():
    class QF(queryfilter.QueryFilter):
        required = True
        variables = {
            't': variable.Variable('t', required=True)
        }

        def apply(self, query):
            return query

    class Q(query.BaseQuery):
        filters = [QF()]

    assert len(Q().find_all_variables()) == 1
    assert Q().find_all_variables()[0].required is True


def test_queryfilter_adds_optional_vars_if_filter_required_and_vars_optional():
    class QF(queryfilter.QueryFilter):
        variables = {
            't': variable.Variable('t', required=True)
        }

        def apply(self, query):
            return query

    class Q(query.BaseQuery):
        filters = [QF()]

    assert len(Q().find_all_variables()) == 1
    assert Q().find_all_variables()[0].required is False


def test_queryfilter_filters_query():
    class QF(queryfilter.QueryFilter):
        variables = {
            't': variable.Variable('t', required=True)
        }

        def apply(self, query, data):
            return {"a": "b"}

    class Q(query.BaseQuery):
        filters = [QF()]

    assert Q().get_es_query({}) == {
        'aggs': {},
        'query': {},
        'size': 0,
        'sort': []
    }
    assert Q().get_es_query({'t': 'x'}) == \
        {'aggs': {},
         'query': {'a': 'b'},
         'size': 0,
         'sort': []
         }
